# type: ignore
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
from pytest import mark, raises

from .url_db import UrlDBError, verify_db


@mark.parametrize(
    "data",
    [
        # valid
        {"d": {"s": ["/"]}},
        # valid
        {"1": {"2": ["/"]}},
        # valid
        {"1d": {"s2": ["/"]}},
        # valid no subdomin
        {"d": {"": ["/"]}},
    ],
)
def test_verify_db(data):
    """test verify_db()"""
    verify_db(data)


@mark.parametrize(
    "data, msg",
    [
        # not a dict
        ([], "Data must be stored in a dict"),
        # empty domain name
        ({"": {"s": ["/"]}}, "Domain must be a string"),
        # subdomain is not a string
        ({"d": {None: ["/"]}}, "Subdomain must be a string"),
        # empty domain entry
        ({"d": {}}, "Invalid domain entry: 'd'"),
        # empty subdomain entry
        ({"d": {"s": []}}, "Invalid subdomain entry: 's' in 'd'"),
        # empty path entry
        ({"d": {"s": [""]}}, "Path must be a string starting with '/'"),
        # upper case value in domain
        ({"D": {"s": ["/"]}}, "Domain must be lower case: 'D'"),
        # upper case value in subdomain
        ({"d": {"S": ["/"]}}, "Subdomain must be lower case: 'S' in 'd'"),
    ],
)
def test_verify_db_failures(data, msg):
    """test verify_db() failures"""
    with raises(UrlDBError, match=msg):
        verify_db(data)



# type: ignore
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
from ffpuppet import BrowserTerminatedError, BrowserTimeoutError, Reason
from pytest import raises

from .browser_wrapper import BrowserArgs, BrowserState
from .explorer import ExplorerMode
from .firefox_wrapper import FirefoxWrapper
from .url import URL


def test_firefox_wrapper_basic(mocker, tmp_path):
    """test FirefoxWrapper basic"""
    args = BrowserArgs(tmp_path / "firefox", 10, 10)
    assert FirefoxWrapper.environment_manager(0, args) is None
    ffp = mocker.patch(
        "site_scout.firefox_wrapper.FFPuppet", autospec=True
    ).return_value
    with FirefoxWrapper(args, tmp_path, None) as browser:
        working_path = browser._working_path
        assert working_path.is_dir()
        assert not browser.debugger()
        # set is_healthy
        ffp.is_healthy.return_value = False
        assert not browser.is_healthy()
        assert ffp.is_healthy.call_count == 1
        # dump coverage should not run when is_healthy is false
        browser.dump_coverage()
        assert ffp.dump_coverage.call_count == 0
        # is_idle should always be true when is_healthy is false
        assert browser.is_idle(idle_limit=1)
        assert ffp.cpu_usage.call_count == 0
        # launch
        browser.launch(URL("a.com"), False, None, False)
        assert ffp.launch.call_count == 1
        assert ffp.clean_up.call_count == 0
        ffp.is_healthy.return_value = True
        # wait wrapper test
        browser.wait(0)
        assert ffp.wait.call_count == 1
        # dump coverage should run when is_healthy is true
        browser.dump_coverage()
        assert ffp.dump_coverage.call_count == 1
        # is_idle should check cpu usage when is_healthy is true
        ffp.cpu_usage.return_value = ((123, 99),)
        assert not browser.is_idle(idle_limit=10)
        assert ffp.cpu_usage.call_count == 1
        # close
        browser.close()
        assert ffp.close.call_count == 1
    assert ffp.clean_up.call_count == 1
    assert not working_path.is_dir()


def test_firefox_wrapper_state(mocker, tmp_path):
    """test FirefoxWrapper.state"""
    ffp = mocker.patch(
        "site_scout.firefox_wrapper.FFPuppet", autospec=True
    ).return_value
    args = BrowserArgs(tmp_path / "firefox", 10, 10)
    with FirefoxWrapper(args, tmp_path, None) as browser:
        ffp.reason = Reason.ALERT
        assert browser.state() == BrowserState.RESULT
        ffp.reason = Reason.WORKER
        assert browser.state() == BrowserState.RESULT
        ffp.reason = Reason.CLOSED
        assert browser.state() == BrowserState.CLOSED
        ffp.reason = Reason.EXITED
        assert browser.state() == BrowserState.EXITED


def test_firefox_wrapper_launch_failure(mocker, tmp_path):
    """test FirefoxWrapper.launch() failure"""
    ffp = mocker.patch(
        "site_scout.firefox_wrapper.FFPuppet", autospec=True
    ).return_value
    args = BrowserArgs(tmp_path / "firefox", 10, 10)
    with FirefoxWrapper(args, tmp_path, None) as browser:
        ffp.launch.side_effect = BrowserTimeoutError()
        with raises(BrowserTimeoutError):
            assert not browser.launch(
                "http://a.com", True, tmp_path, raise_on_failure=True
            )
        assert ffp.save_logs.call_count == 0
        assert ffp.clean_up.call_count == 1

        ffp.reset_mock()
        ffp.launch.side_effect = BrowserTerminatedError()
        assert not browser.launch(
            "http://a.com", True, tmp_path, raise_on_failure=False
        )
        assert ffp.save_logs.call_count == 0
        assert ffp.clean_up.call_count == 1

        ffp.reset_mock()
        ffp.launch.side_effect = BrowserTerminatedError()
        with raises(BrowserTerminatedError):
            browser.launch("http://a.com", True, tmp_path, raise_on_failure=True)
        assert ffp.save_logs.call_count == 1
        assert ffp.clean_up.call_count == 1


def test_firefox_wrapper_save_report(mocker, tmp_path):
    """test FirefoxWrapper.save_report()"""
    ffp = mocker.patch(
        "site_scout.firefox_wrapper.FFPuppet", autospec=True
    ).return_value
    prefs = tmp_path / "original_prefs.js"
    prefs.write_text("foo")
    args = BrowserArgs(tmp_path / "firefox", 10, 10, prefs_file=prefs)
    with FirefoxWrapper(args, tmp_path, None) as browser:
        browser.save_report(tmp_path)
        assert ffp.save_logs.call_count == 1
    assert (tmp_path / "prefs.js").is_file()


def test_firefox_wrapper_create_explorer(mocker, tmp_path):
    """test FirefoxWrapper.create_explorer()"""
    explorer = mocker.patch("site_scout.firefox_wrapper.Explorer", autospec=True)
    ffp = mocker.patch("site_scout.firefox_wrapper.FFPuppet", autospec=True)
    ffp.return_value.marionette = 12345
    args = BrowserArgs(tmp_path / "firefox", 10, 10)
    with FirefoxWrapper(args, tmp_path, None) as browser:
        assert browser.create_explorer(URL("a.com"), 10, mode=ExplorerMode.ALL, pause=5)
    explorer.assert_called_with(
        args.binary,
        12345,
        "http://a.com/",
        mode=ExplorerMode.ALL,
        load_wait=10,
        pause=5,
    )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from hashlib import sha1
from logging import WARNING, getLogger
from re import compile as re_compile
from string import punctuation
from urllib.parse import quote, urlsplit

from tldextract import extract

# suppress output from filelock (used by tldextract)
getLogger("filelock").setLevel(WARNING)


class URLParseError(Exception):
    """Provided input cannot be converted to a URL."""


class URL:
    """URL components."""

    ALLOWED_SCHEMES = frozenset(("http", "https"))
    VALID_DOMAIN = re_compile(r"[a-zA-Z0-9:.-]+")
    VALID_SUBDOMAIN = re_compile(r"[a-zA-Z0-9_.-]+")

    __slots__ = ("_uid", "alias", "domain", "path", "scheme", "subdomain")

    def __init__(
        self,
        domain: str,
        subdomain: str = "",
        path: str = "/",
        scheme: str = "http",
    ) -> None:
        self.alias: str | None = None
        self.domain = domain
        self.path = path
        self.scheme = scheme
        self.subdomain = subdomain
        self._uid: str | None = None

    def __str__(self) -> str:
        if not self.subdomain:
            return f"{self.scheme}://{self.domain}{self.path}"
        return f"{self.scheme}://{self.subdomain}.{self.domain}{self.path}"

    @classmethod
    def create(
        cls,
        domain: str,
        subdomain: str = "",
        path: str = "/",
        scheme: str = "http",
    ) -> URL:
        """Sanitize, verify data and create a URL if possible.

        Args:
            domain: Domain.
            subdomain: Subdomain.
            path: Path, must begin with '/'.
            scheme: Scheme.

        Returns:
            URL object if input is valid otherwise None.
        """
        scheme = scheme.lower()
        if scheme not in cls.ALLOWED_SCHEMES:
            raise URLParseError(f"Unsupported scheme '{scheme}'")

        try:
            # use idna to encode domain with non ascii characters
            domain = domain.lower().encode("idna").decode("ascii")
        except UnicodeError:
            raise URLParseError(f"Invalid domain '{domain}'") from None
        if cls.VALID_DOMAIN.fullmatch(domain) is None:
            raise URLParseError(f"Invalid domain '{domain}'")

        if subdomain:
            try:
                subdomain = subdomain.lower().encode("idna").decode("ascii")
            except UnicodeError:
                raise URLParseError(f"Invalid subdomain '{subdomain}'") from None
            if cls.VALID_SUBDOMAIN.fullmatch(subdomain) is None:
                raise URLParseError(f"Invalid subdomain '{subdomain}'")

        if not path.startswith("/"):
            raise URLParseError("Path must begin with '/'")
        # percent encode non ascii characters in path if needed
        if not path.isascii():
            path = quote(path, safe=punctuation)

        return cls(domain, subdomain=subdomain, path=path, scheme=scheme)

    @classmethod
    def parse(cls, url: str) -> URL:
        """Parse URL from a given string. Only URLs with a valid domain and tld are
        supported. IP addresses, `about:` pages, etc, are not supported.

        Args:
            url: Input to parse.

        Returns:
            URL object if string is successfully parsed.
        """
        url = url.strip()
        if "://" not in url:
            url = f"http://{url}"
        try:
            parsed = urlsplit(url, allow_fragments=False)
        except ValueError:
            raise URLParseError("Invalid URL") from None

        # strip credentials
        if "@" in parsed.netloc:
            netloc = parsed.netloc.rsplit("@")[-1]
        else:
            netloc = parsed.netloc

        # parse domain, subdomain and tld
        udi = extract(netloc)
        if not udi.domain:
            raise URLParseError("Missing domain")
        if not udi.suffix:
            raise URLParseError("Missing suffix")
        if not udi.subdomain and netloc[0] == ".":
            raise URLParseError("Invalid subdomain")

        # parse port
        if ":" in netloc:
            try:
                port = int(netloc.rsplit(":", maxsplit=1)[-1])
            except ValueError:
                raise URLParseError("Port must be a number") from None
            if port < 1 or port > 65535:
                raise URLParseError("Invalid port number")
            domain = f"{udi.domain}.{udi.suffix}:{port}"
        else:
            domain = f"{udi.domain}.{udi.suffix}"

        # parse path
        path = parsed.path or "/"
        if parsed.query:
            path = f"{path}?{parsed.query}"

        # sanitize parsed data and create URL
        return URL.create(
            domain,
            subdomain=udi.subdomain,
            path=path,
            scheme=parsed.scheme,
        )

    @property
    def uid(self) -> str:
        """Unique ID.

        Args:
            None

        Returns:
            Unique ID.
        """
        # this does NOT need to be cryptographically secure
        # it needs to be filesystem safe and *somewhat* unique
        if not self._uid:
            self._uid = sha1(
                str(self).encode(errors="replace"), usedforsecurity=False
            ).hexdigest()
        return self._uid



# type: ignore
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
from logging import NOTSET, disable

from ffpuppet import BrowserTerminatedError
from pytest import mark, raises

from .main import generate_prefs, load_jsonl, load_yml, main, scan_input
from .url_db import UrlDBError


@mark.parametrize(
    "args",
    [
        # only required args (use yml file)
        [],
        # log-level
        ["--log-level", "DEBUG"],
        # url
        ["-u", "http://mozilla.com"],
        # disable logging
        ["--disable-logging", "-u", "http://private-url.com"],
    ],
)
def test_main_01(caplog, capsys, mocker, tmp_path, args):
    """test main()"""
    browser = mocker.patch("site_scout.main.FirefoxWrapper", autospec=True)
    browser.return_value.is_healthy.return_value = False
    fake_bin = tmp_path / "fake_browser_bin"
    fake_bin.touch()
    default_args = [str(fake_bin), "-b", "firefox"]
    # -i or -u is required (create yml file)
    if "-u" not in args:
        # test yml loading path
        yml_file = tmp_path / "data.yml"
        yml_file.write_text("{'d':{'s':['/']}}")
        # test jsonl loading path
        jsonl_file = tmp_path / "data.jsonl"
        jsonl_file.write_text('{"a": "b"}\n')
        default_args.extend(["-i", str(yml_file), str(jsonl_file)])
    try:
        assert main(default_args + args) == 0
    finally:
        # always re-enable logging
        disable(level=NOTSET)
    assert browser.return_value.launch.call_count == 1
    if "--disable-logging" in args:
        # verify we are not sending any unexpected data to the console
        console = capsys.readouterr()
        assert not console.err
        assert not console.out
        assert "Running with logging disabled..." in caplog.text
        assert "Starting Site Scout..." not in caplog.text
        assert "Done." not in caplog.text
        assert "private-url.com" not in caplog.text
        assert len(caplog.text.splitlines()) == 1
    else:
        assert "Starting Site Scout..." in caplog.text
        assert "Done." in caplog.text
        assert "Running with logging disabled..." not in caplog.text
        assert len(caplog.text.splitlines()) > 1


@mark.parametrize(
    "yml_data",
    [
        # invalid yml data
        "\0",
        # invalid data in yml
        "{'d':[]}",
    ],
)
def test_main_load_yml_invalid(caplog, mocker, tmp_path, yml_data):
    """test main()"""
    browser = mocker.patch("site_scout.main.FirefoxWrapper", autospec=True)
    browser.return_value.is_healthy.return_value = False
    fake_bin = tmp_path / "fake_browser_bin"
    fake_bin.touch()
    yml_file = tmp_path / "data.yml"
    yml_file.write_text(yml_data)
    assert main([str(fake_bin), "-b", "firefox", "-i", str(yml_file)]) == 0
    assert "Starting Site Scout..." in caplog.text
    assert "Done." in caplog.text
    assert "Running with logging disabled..." not in caplog.text
    assert len(caplog.text.splitlines()) > 1


@mark.parametrize(
    "exception, exit_code",
    [
        # user abort
        (KeyboardInterrupt(), 0),
        # browser failed to launch
        (BrowserTerminatedError(), 1),
    ],
)
def test_main_02(mocker, tmp_path, exception, exit_code):
    """test main() exceptions"""
    mocker.patch("site_scout.main.SiteScout", autospec=True, side_effect=exception)
    fake_bin = tmp_path / "fake_browser_bin"
    fake_bin.touch()
    assert main([str(fake_bin), "-u", "a.b.c"]) == exit_code


def test_generate_prefs_01(tmp_path):
    """test generate_prefs()"""
    assert generate_prefs(dst=tmp_path).is_file()


def test_load_yml_simple(tmp_path):
    """test load_yml() with valid data"""
    (tmp_path / "sites.yml").write_text("{'d':{'s':['/']}}")
    assert len(load_yml(tmp_path / "sites.yml")) == 1


def test_load_yml_invalid_data(tmp_path):
    """test load_yml() with invalid data"""
    (tmp_path / "test.yml").write_text("\0")
    with raises(UrlDBError, match="Invalid yml"):
        load_yml(tmp_path / "test.yml")


def test_load_jsonl_simple(tmp_path):
    """test load_jsonl() with valid data"""
    (tmp_path / "sites.jsonl").write_text(
        '{"a.com": "423-456-7853"}\n'
        '{"b.com": "333-422-2222"}\n'
        '{"c.com": "123-456-7890"}\n'
        '{"d.com": ""}\n'
        '{"e.com": null}\n'
    )
    results = tuple(load_jsonl(tmp_path / "sites.jsonl"))
    assert len(results) == 5


@mark.parametrize(
    "data",
    [
        # empty file
        b"",
        # invalid chars ' should be "
        b"{'url': 'alias'}\n",
        # empty entry
        b"{}\n",
        # empty lines
        b"\n\n\n",
        # invalid data
        b'{"a": []}\n',
        # invalid data
        b'{"a"}\n',
        # invalid data
        b"[1]\n",
    ],
)
def test_load_jsonl_invalid_data(tmp_path, data):
    """test load_jsonl() with invalid data"""
    (tmp_path / "test.yml").write_bytes(data)
    assert not any(load_jsonl(tmp_path / "test.yml"))


def test_scan_input(tmp_path):
    """test scan_input()"""
    # no paths
    assert not any(scan_input([], ".foo"))
    # empty directory
    assert not any(scan_input([tmp_path], ".foo"))
    # single file
    (tmp_path / "test.yml").touch()
    assert any(scan_input([(tmp_path / "test.yml")], ".yml"))
    # single file in directory
    assert sum(1 for _ in scan_input([tmp_path], ".yml")) == 1
    # no suffix match
    assert not any(scan_input([tmp_path], ".foo"))
    # multiple files in directory
    (tmp_path / "test.jsonl").touch()
    results = tuple(scan_input([tmp_path], ".yml"))
    assert len(results) == 1
    assert results[0].suffix == ".yml"



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum, auto
from logging import getLogger
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path

    from .explorer import Explorer, ExplorerMode
    from .url import URL


LOG = getLogger(__name__)


@dataclass(eq=False, frozen=True)
class BrowserArgs:
    """Arguments specific to browser management.

    Attributes:
        binary: Browser binary.
        launch_timeout: Number of seconds to wait for browser to launch.
        memory_limit: Browser memory limit in bytes.
        debugger: Debugger to use.
        display_mode: Display mode to use.
        prefs_file: Custom prefs.js file to use.
        profile: Custom profile directory to use. TODO: UNUSED, remove?
    """

    binary: Path
    launch_timeout: int
    memory_limit: int
    debugger: str | None = None
    display_mode: str | None = None
    prefs_file: Path | None = None
    profile: Path | None = None


class BrowserState(Enum):
    """Possible browser states when complete."""

    RESULT = auto()
    CLOSED = auto()
    EXITED = auto()


class BrowserWrapper(ABC):
    """BrowserWrapper is used to add support for specific browsers."""

    __slots__ = ("args",)

    # pylint: disable=unused-argument
    def __init__(
        self, args: BrowserArgs, working_path: Path, env_mgr: EnvironmentManager | None
    ) -> None:
        self.args = args

    def __enter__(self) -> BrowserWrapper:
        return self

    def __exit__(self, *exc: object) -> None:
        self.cleanup()

    @abstractmethod
    def cleanup(self) -> None:
        """Cleanup method. Remove all data at that is no longer required.

        Args:
            None

        Returns:
            None.
        """

    @abstractmethod
    def close(self) -> None:
        """Close target.

        Args:
            None

        Returns:
            None.
        """

    @abstractmethod
    def create_explorer(
        self, url: URL, load_wait: int, mode: ExplorerMode, pause: int = 0
    ) -> Explorer:
        """Create Explorer.

        Args:
            url: URL to navigate to.
            load_wait: Maximum number of seconds to wait for page load.
            mode: Operational mode.
            pause: Number of seconds to wait after page has loaded before continuing.

        Return:
            Explorer object.
        """

    @abstractmethod
    def debugger(self) -> bool:
        """Check if an external debugger is in use.

        Args:
            None

        Returns:
            True if a debugger is in use otherwise False.
        """

    @abstractmethod
    def dump_coverage(self, timeout: int = 0) -> None:
        """Trigger coverage data dump.

        Args:
            timeout: Amount of time to wait for data to be written.

        Returns:
            None.
        """

    @staticmethod
    @abstractmethod
    def environment_manager(
        instance_limit: int,
        browser_args: BrowserArgs,
    ) -> EnvironmentManager | None:
        """Create a custom environment manager instance.

        Args:
            instance_limit: Maximum number of browser instances to run in parallel.
            browser_args: Arguments passed to the browser.

        Returns:
            An environment manager object or None.
        """

    @abstractmethod
    def is_healthy(self) -> bool:
        """Check the browser is in a good state.

        Args:
            None

        Returns:
            True if the browser is running and determined to be
            in a valid functioning state otherwise False.
        """

    @abstractmethod
    def is_idle(self, idle_limit: int) -> bool:
        """Check if the browser is in an idle state.

        Args:
            idle_limit: Maximum CPU process usage to be considered idle.

        Returns:
            True if the browser is idle otherwise False.
        """

    @abstractmethod
    def launch(
        self,
        url: str,
        explore: bool,
        log_path: Path | None,
        raise_on_failure: bool,
    ) -> bool:
        """Launch the browser.

        Args:
            url: URL to load.
            explore: Use an Explorer to interact with the page.
            log_path: Location to save unexpected launch error logs.
            raise_on_failure: Raise or suppress launch a failure.

        Returns:
            True if the browser launched successfully otherwise False.
        """

    @abstractmethod
    def save_report(self, dst: Path) -> None:
        """Save logs and other related files to specified location.

        Args:
            dst: Location to save report.

        Returns:
            None.
        """

    @abstractmethod
    def state(self) -> BrowserState:
        """Browser state when complete. This is only intended to be used when the
        browser has executed and is no longer running. For example it indicates if the
        browser was closed, if it exited or if it crashed.

        Args:
            None.

        Returns:
            Browser state.
        """

    @abstractmethod
    def wait(self, timeout: float) -> bool:
        """Wait for a given amount of time while the browser is running.

        Args:
            timeout: Number of seconds to wait.

        Return:
            True if the browser exited before the timeout elapsed otherwise False.
        """


class EnvironmentManager(ABC):
    """EnvironmentManager is used to configure and control required components that are
    required to run browsers, for example, emulators."""

    @abstractmethod
    def cleanup(self) -> None:
        """Perform any necessary cleanup operations"""



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from logging import getLogger
from pathlib import Path
from shutil import copyfile, rmtree
from tempfile import mkdtemp
from time import strftime
from typing import TYPE_CHECKING

from ffpuppet import (
    BrowserTimeoutError,
    Debugger,
    DisplayMode,
    FFPuppet,
    LaunchError,
    Reason,
)

from .browser_wrapper import (
    BrowserArgs,
    BrowserState,
    BrowserWrapper,
)
from .explorer import Explorer, ExplorerMode

if TYPE_CHECKING:
    from .url import URL


LOG = getLogger(__name__)


class FirefoxWrapper(BrowserWrapper):
    """FirefoxWrapper is used to add support for Firefox."""

    __slots__ = ("_debugger", "_ffp", "_working_path")

    def __init__(self, args: BrowserArgs, working_path: Path, env_mgr: None) -> None:
        super().__init__(args, working_path, env_mgr)
        self._working_path = Path(mkdtemp(dir=working_path))
        # TODO: generate prefs if needed
        self._debugger = args.debugger or Debugger.NONE.name
        display_mode = args.display_mode or DisplayMode.DEFAULT.name
        self._ffp = FFPuppet(
            debugger=Debugger[self._debugger.upper()],
            display_mode=DisplayMode[display_mode.upper()],
            working_path=str(self._working_path),
        )

    def cleanup(self) -> None:
        self._ffp.clean_up()
        rmtree(self._working_path, ignore_errors=True)

    def close(self) -> None:
        self._ffp.close()

    def create_explorer(
        self, url: URL, load_wait: int, mode: ExplorerMode, pause: int = 0
    ) -> Explorer:
        assert self._ffp.marionette is not None
        # this can raise RuntimeError
        return Explorer(
            self.args.binary,
            self._ffp.marionette,
            str(url),
            mode=mode,
            load_wait=load_wait,
            pause=pause,
        )

    def debugger(self) -> bool:
        return self._debugger.upper() != Debugger.NONE.name

    def dump_coverage(self, timeout: int = 15) -> None:
        if self._ffp.is_healthy():
            self._ffp.dump_coverage(timeout=timeout)

    @staticmethod
    def environment_manager(instance_limit: int, browser_args: BrowserArgs) -> None:
        return None

    def is_healthy(self) -> bool:
        return self._ffp.is_healthy()

    def is_idle(self, idle_limit: int) -> bool:
        if self._ffp.is_healthy():
            # check that no processes are over the limit
            return not any(pct > idle_limit for _, pct in self._ffp.cpu_usage())
        # not running
        return True

    def launch(
        self,
        url: str,
        explore: bool,
        log_path: Path | None,
        raise_on_failure: bool,
    ) -> bool:
        success = False
        try:
            self._ffp.launch(
                self.args.binary,
                env_mod={"MOZ_CRASHREPORTER_SHUTDOWN": "1"},
                location=None if explore else str(url),
                launch_timeout=self.args.launch_timeout,
                log_limit=2 * 1024 * 1024,
                marionette=0 if explore else None,
                memory_limit=self.args.memory_limit,
                prefs_js=self.args.prefs_file,
            )
            success = True
        except LaunchError as exc:
            is_failure = not isinstance(exc, BrowserTimeoutError)
            LOG.warning("Browser launch %s...", "failure" if is_failure else "timeout")
            if raise_on_failure:
                # save failure
                if is_failure and log_path is not None:
                    self._ffp.close()
                    dst = log_path / strftime("%Y%m%d-%H%M%S-launch-failure")
                    self.save_report(dst)
                    LOG.warning("Logs saved '%s'", dst)
                raise
        finally:
            # cleanup if launch was unsuccessful
            if not success:
                self._ffp.clean_up()
        return success

    def save_report(self, dst: Path) -> None:
        self._ffp.save_logs(dst)
        # save prefs file
        if self.args.prefs_file is not None:
            copyfile(self.args.prefs_file, dst / "prefs.js")

    def state(self) -> BrowserState:
        assert self._ffp.reason is not None
        if self._ffp.reason == Reason.EXITED:
            return BrowserState.EXITED
        if self._ffp.reason in {Reason.ALERT, Reason.WORKER}:
            return BrowserState.RESULT
        return BrowserState.CLOSED

    def wait(self, timeout: float) -> bool:
        return self._ffp.wait(timeout=timeout)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from contextlib import suppress
from json import JSONDecodeError, loads
from logging import DEBUG, basicConfig, disable, getLogger
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING

from ffpuppet import LaunchError
from prefpicker import PrefPicker
from yaml import load

try:
    from yaml import CSafeLoader as SafeLoader
except ImportError:  # pragma: no cover
    from yaml import SafeLoader  # type: ignore

from yaml.parser import ParserError
from yaml.reader import ReaderError
from yaml.scanner import ScannerError

from .args import parse_args
from .browser_wrapper import BrowserArgs
from .fenix_wrapper import FenixWrapper
from .firefox_wrapper import FirefoxWrapper
from .site_scout import SiteScout
from .url_db import UrlDB, UrlDBError, verify_db

if TYPE_CHECKING:
    from collections.abc import Generator, Iterable

    from .browser_wrapper import BrowserWrapper

LOG = getLogger(__name__)


def init_logging(level: int, disable_logging: bool) -> None:
    """Initialize logging.

    Arguments:
        level: logging verbosity level.
        disable_logging: Disable log output.

    Returns:
        None
    """
    if level == DEBUG:
        log_fmt = (
            "%(asctime)s.%(msecs)03d %(levelname).1s %(threadName)s %(name)s | "
            "%(message)s"
        )
    else:
        log_fmt = "[%(asctime)s] %(message)s"
    basicConfig(format=log_fmt, datefmt="%H:%M:%S", level=level)
    if disable_logging:
        LOG.warning("Running with logging disabled...")
        disable()


def generate_prefs(dst: Path | None = None, variant: str = "a11y") -> Path:
    """Generate prefs.js file via PrefPicker.

    Arguments:
        variant: Variant to use.

    Returns:
        pref.js file.
    """
    with NamedTemporaryFile(prefix="scout-prefs-", suffix=".js", dir=dst) as tmp_fp:
        prefs = Path(tmp_fp.name)
    picker = PrefPicker.load_template(
        Path(__file__).parent / "resources" / "prefpicker.yml"
    )
    picker.create_prefsjs(prefs, variant=variant)
    return prefs


def load_jsonl(src: Path) -> Generator[tuple[str, str | None]]:
    """Load data from a jsonl file.

    The expected format is:
        {"url1": "alias1"}\n
        {"url2": "alias2"}\n
        {"url3": null}\n
        ...

    Arguments:
        src: File to load data from.

    Yields:
        URL and alias.
    """
    LOG.debug("loading (jsonl) '%s'", src.resolve())
    with src.open("r") as in_fp:
        line_no = 0
        try:
            for line_no, line in enumerate(in_fp, start=1):
                line = line.strip()
                # ignore blank lines
                if not line:
                    continue
                result = loads(line)
                # there should be exactly one entry
                if not isinstance(result, dict) or len(result) != 1:
                    LOG.error("Invalid format: '%s:%d'", src.resolve(), line_no)
                    break
                url, alias = result.popitem()
                if alias is not None and not isinstance(alias, str):
                    LOG.error("Invalid alias: '%s:%d'", src.resolve(), line_no)
                    break
                yield url, alias
        except JSONDecodeError as exc:
            LOG.error("Invalid data: '%s:%d' (%s)", src.resolve(), line_no, exc)


def load_yml(src: Path) -> UrlDB:
    """Load data from a yml file.

    Arguments:
        src: File to load data from.

    Returns:
        URL data.
    """
    LOG.debug("loading (yml) '%s'", src.resolve())
    with (
        src.open("r") as in_fp,
        suppress(ParserError, ReaderError, ScannerError),
    ):
        return verify_db(load(in_fp, Loader=SafeLoader))
    raise UrlDBError("Invalid yml") from None


def scan_input(src: Iterable[Path], suffix: str) -> Generator[Path]:
    """Scan provided locations for input files.

    Arguments:
        src: Paths to scan.
        suffix: Only files with this suffix will be included.

    Yields:
        Data files to load.
    """
    assert suffix.startswith(".")
    for entry in src:
        if entry.is_dir():
            yield from entry.glob(f"*{suffix}")
        elif entry.suffix == suffix:
            yield entry


def main(argv: list[str] | None = None) -> int:
    """Main function"""
    args = parse_args(argv)
    assert any(args.input) != any(args.url)
    init_logging(args.log_level, args.disable_logging)

    # available browser wrappers
    wrappers: dict[str, type[BrowserWrapper]] = {
        "firefox": FirefoxWrapper,
        "fenix": FenixWrapper,
    }
    LOG.debug("browser to use '%s'", args.browser)

    tmp_prefs = False
    try:
        # generate prefs.js file if one is not provided
        if args.prefs is None:
            # TODO: move this to the wrapper
            args.prefs = generate_prefs()
            tmp_prefs = True

        LOG.info("Starting Site Scout...")
        with SiteScout(
            wrappers[args.browser],
            BrowserArgs(
                args.binary.resolve(),
                args.launch_timeout,
                args.memory_limit,
                prefs_file=args.prefs,
                profile=args.profile,
                debugger=args.debugger,
                display_mode=args.display,
            ),
            fuzzmanager=args.fuzzmanager,
            coverage=args.coverage,
            explore=args.explore,
            omit_urls=args.omit_urls,
        ) as scout:
            LOG.info("Loading URLs...")
            # load jsonl files
            for src in scan_input(args.input, ".jsonl"):
                for in_url, alias in load_jsonl(src):
                    scout.load_str(in_url, alias=alias)
            # load yml files
            for src in scan_input(args.input, ".yml"):
                try:
                    scout.load_db(load_yml(src))
                except UrlDBError as exc:  # noqa: PERF203
                    LOG.warning("Load failure - %s (ignored: %s)", exc, src.resolve())
            # load urls from command line
            for in_url in args.url:
                scout.load_str(in_url)
            # don't randomize urls passed on the command line
            scout.schedule_urls(
                url_limit=args.url_limit,
                randomize=any(args.input),
                visits=args.visits,
            )
            scout.run(
                args.output_path,
                args.time_limit,
                instance_limit=args.jobs,
                status_report=args.status_report,
                result_limit=args.result_limit,
                runtime_limit=args.runtime_limit,
            )
    except KeyboardInterrupt:
        LOG.warning("Aborting...")

    except LaunchError as exc:
        LOG.error(str(exc))
        return 1

    finally:
        if tmp_prefs:
            args.prefs.unlink()
        LOG.info("Done.")

    return 0



# type: ignore
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
from platform import system

from pytest import mark, raises

from .args import TIME_LIMIT_DEBUG, TIME_LIMIT_DEFAULT, TIME_LIMIT_EXPLORE, parse_args


@mark.parametrize(
    "file_name, browser",
    [
        ("firefox", "firefox"),
        ("target.apk", "fenix"),
    ],
)
def test_parse_args_basic(tmp_path, file_name, browser):
    """test parse_args()"""
    binary = tmp_path / file_name
    binary.touch()
    args = parse_args([str(binary), "-i", str(binary)])
    assert args.browser == browser


def test_parse_args_checks(capsys, tmp_path):
    """test parse_args() checks"""
    dummy_file = tmp_path / "dummy_file"
    dummy_file.touch()

    with raises(SystemExit):
        parse_args(["missing", "-i", str(dummy_file)])
    assert "binary does not exist: 'missing'" in capsys.readouterr()[-1]

    with raises(SystemExit):
        parse_args([str(dummy_file), "-i", "missing"])
    assert "-i/--input does not exist: 'missing'" in capsys.readouterr()[-1]

    with raises(SystemExit):
        parse_args([str(dummy_file), "-i", str(dummy_file), "-o", "missing"])
    assert "-o/--output-path does not exist: 'missing'" in capsys.readouterr()[-1]

    with raises(SystemExit):
        parse_args([str(dummy_file), "-i", str(dummy_file), "--prefs", "missing"])
    assert "--prefs does not exist: 'missing'" in capsys.readouterr()[-1]

    with raises(SystemExit):
        parse_args([str(dummy_file), "-i", str(dummy_file), "--jobs", "0"])
    assert "--jobs must be >= 1" in capsys.readouterr()[-1]

    with raises(SystemExit):
        parse_args([str(dummy_file), "-i", str(dummy_file), "--memory-limit", "-1"])
    assert "--memory-limit must be >= 0" in capsys.readouterr()[-1]

    with raises(SystemExit):
        parse_args([str(dummy_file), "-i", str(dummy_file), "--time-limit", "-1"])
    assert "--time-limit must be > 0" in capsys.readouterr()[-1]

    with raises(SystemExit):
        parse_args(
            [str(dummy_file), "-i", str(dummy_file), "--explore", "--browser", "fenix"]
        )
    assert "-explore not supported with selected browser" in capsys.readouterr()[-1]


def test_parse_args_debugger(mocker, capsys, tmp_path):
    """test parse_args() rr/pernosco checks"""
    mocker.patch("site_scout.args.system", autospec=True, return_value="Linux")
    fake_which = mocker.patch("site_scout.args.which", autospec=True)
    dummy_file = tmp_path / "dummy_file"
    dummy_file.touch()

    with raises(SystemExit):
        parse_args([str(dummy_file), "-u", "foo", "--rr", "--fuzzmanager"])
    assert "rr not supported with --fuzzmanager" in capsys.readouterr()[-1]

    fake_which.return_value = None
    with raises(SystemExit):
        parse_args([str(dummy_file), "-u", "foo", "--rr"])
    assert "rr is not installed" in capsys.readouterr()[-1]

    fake_which.return_value = "rr"
    mocker.patch("site_scout.args.Path.read_bytes", autospec=True, return_value=b"99")
    with raises(SystemExit):
        parse_args([str(dummy_file), "-u", "foo", "--rr"])
    assert "perf_event_paranoid <= 1, but it is 99" in capsys.readouterr()[-1]


@mark.skipif(system() != "Linux", reason="Only supported on Linux")
def test_parse_args_coverage(capsys, mocker, tmp_path):
    """test parse_args() - coverage"""
    dummy_file = tmp_path / "dummy_file"
    dummy_file.touch()
    fake_getenv = mocker.patch("site_scout.args.getenv", autospec=True)

    assert parse_args([str(dummy_file), "-i", str(dummy_file), "--coverage"])

    fake_getenv.side_effect = ("", "", "")
    with raises(SystemExit):
        assert parse_args([str(dummy_file), "-i", str(dummy_file), "--coverage"])
    assert (
        "error: GCOV_PREFIX_STRIP must be set to use --coverage"
        in capsys.readouterr()[-1]
    )

    fake_getenv.side_effect = ("", "", "/", "")
    with raises(SystemExit):
        assert parse_args([str(dummy_file), "-i", str(dummy_file), "--coverage"])
    assert "error: GCOV_PREFIX must be set to use --coverage" in capsys.readouterr()[-1]

    fake_getenv.side_effect = ("", "", "/", "/")
    with raises(SystemExit):
        assert parse_args(
            [str(dummy_file), "-i", str(dummy_file), "--coverage", "--jobs", "2"]
        )
    assert (
        "error: Parallel jobs not supported with --coverage" in capsys.readouterr()[-1]
    )


@mark.parametrize(
    "extra_args, expected_time",
    [
        ([], TIME_LIMIT_DEFAULT),
        (["--explore"], TIME_LIMIT_EXPLORE),
        (["--pernosco"], TIME_LIMIT_DEBUG),
        (["--time-limit", "111"], 111),
    ],
)
def test_args_time_limit(mocker, tmp_path, extra_args, expected_time):
    """test parse_args() set time limit"""
    mocker.patch("site_scout.args.Path.read_bytes", autospec=True, return_value=b"0")
    mocker.patch("site_scout.args.system", autospec=True, return_value="Linux")
    mocker.patch("site_scout.args.which", autospec=True)

    dummy_file = tmp_path / "dummy_file"
    dummy_file.touch()
    args = parse_args([str(dummy_file), "-i", str(dummy_file), *extra_args])
    assert args.time_limit == expected_time



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Any, NewType, cast

# Note: Python 3.10+ use TypeAlias
UrlDB = NewType("UrlDB", dict[str, dict[str, list[str]]])


class UrlDBError(Exception):
    """Raised when UrlDB related issues are detected."""


def verify_db(data: Any) -> UrlDB:
    """Verify UrlDB structure and data.

    Args:
        data: Data to check.

    Return:
        UrlDB.
    """
    if not isinstance(data, dict):
        raise UrlDBError("Data must be stored in a dict")
    # check domains
    for domain, subdomains in data.items():
        if not isinstance(domain, str) or not domain:
            raise UrlDBError("Domain must be a string")
        if not isinstance(subdomains, dict) or not subdomains:
            raise UrlDBError(f"Invalid domain entry: '{domain}'")
        if any(x.isupper() for x in domain):
            raise UrlDBError(f"Domain must be lower case: '{domain}'")
        # check subdomains
        for subdomain, paths in subdomains.items():
            if not isinstance(subdomain, str):
                raise UrlDBError("Subdomain must be a string")
            if not isinstance(paths, list) or not paths:
                raise UrlDBError(
                    f"Invalid subdomain entry: '{subdomain}' in '{domain}'"
                )
            if any(x.isupper() for x in subdomain):
                raise UrlDBError(
                    f"Subdomain must be lower case: '{subdomain}' in '{domain}'"
                )
            # check paths
            for path in paths:
                if not isinstance(path, str) or not path.startswith("/"):
                    raise UrlDBError("Path must be a string starting with '/'")
    return cast("UrlDB", data)



# type: ignore
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access

from pytest import mark, raises

from .url import URL, URLParseError


def test_url_str():
    """test URL.__str__()"""
    assert str(URL("a.c", subdomain="b", path="/d")) == "http://b.a.c/d"


@mark.parametrize(
    "domain, subdomain, path, scheme, expected",
    [
        # with a subdomain
        ("a.c", "b", "/d", "http", "http://b.a.c/d"),
        # without a subdomain
        ("a.c", None, "/d", "http", "http://a.c/d"),
        # more complex
        ("a.c:1337", "x.y", "/1/2/3.html", "https", "https://x.y.a.c:1337/1/2/3.html"),
        # normalize case (domain and subdomain only)
        ("aB.cD", "Ab", "/Ef", "HTTP", "http://ab.ab.cd/Ef"),
        # normalize case (missing subdomain)
        ("aB.cD", None, "/Ef", "HTTP", "http://ab.cd/Ef"),
        # idna encode domain
        ("é.c", "ò", "/", "http", "http://xn--jda.xn--9ca.c/"),
        # percent encode non ascii chars path
        ("a.c", None, "/ñ%41=1", "http", "http://a.c/%C3%B1%41=1"),
        # previously encoded path
        ("a.c", None, "/%C3%B1o", "http", "http://a.c/%C3%B1o"),
    ],
)
def test_url_create(domain, subdomain, path, scheme, expected):
    """test URL.create()"""
    url = URL.create(domain, subdomain=subdomain, path=path, scheme=scheme)
    assert str(url) == expected
    assert url.uid


@mark.parametrize(
    "domain, subdomain, path, scheme, msg",
    [
        # bad domain
        ("*", None, "/", "http", "Invalid domain '*'"),
        # bad domain
        ("", None, "/", "http", "Invalid domain ''"),
        # bad domain
        ("b a.d", None, "/", "http", "Invalid domain 'b a.d'"),
        # bad domain
        ("b_a.d", None, "/", "http", "Invalid domain 'b_a.d'"),
        # bad domain
        (".a.c", None, "/", "http", "Invalid domain '.a.c'"),
        # bad subdomain
        ("a.c", "$.foo", "/", "http", r"Invalid subdomain '\$.foo'"),
        # bad subdomain
        ("a.c", ".a", "/", "http", "Invalid subdomain '.a'"),
        # bad subdomain
        ("a.c", "..a", "/", "http", "Invalid subdomain '..a'"),
        # bad subdomain
        ("d.c", "b a", "/", "http", "Invalid subdomain 'b a'"),
        # bad subdomain
        ("d.c", "b:a", "/", "http", "Invalid subdomain 'b:a'"),
        # bad path
        ("a.c", None, "", "http", "Path must begin with '/'"),
        # bad path
        ("a.c", None, "foo", "http", "Path must begin with '/'"),
        # bad scheme
        ("a.c", None, "/", "foo", "Unsupported scheme 'foo'"),
    ],
)
def test_url_create_failures(domain, subdomain, path, scheme, msg):
    """test URL.create() - failures"""
    with raises(URLParseError, match=msg):
        URL.create(domain, subdomain=subdomain, path=path, scheme=scheme)


@mark.parametrize(
    "url_str, expected",
    [
        # http scheme, domain and tld
        ("http://domain.com/", "http://domain.com/"),
        # https scheme, domain and tld
        ("https://domain.com/", "https://domain.com/"),
        # http scheme, domain, subdomain and tld
        ("http://sub.domain.com/", "http://sub.domain.com/"),
        # https scheme, domain, subdomain and tld
        ("https://sub.domain.com/", "https://sub.domain.com/"),
        # subdomain, domain and tld (missing scheme)
        ("sub.domain.com", "http://sub.domain.com/"),
        # domain and tld (missing scheme)
        ("domain.com", "http://domain.com/"),
        # trailing space
        ("domain.com ", "http://domain.com/"),
        # trailing space after path
        ("domain.com/ ", "http://domain.com/"),
        # trailing space after path
        ("domain.com/test ", "http://domain.com/test"),
        # contains credentials
        ("user:pass@domain.com", "http://domain.com/"),
        # multipart tld
        ("domain.co.uk", "http://domain.co.uk/"),
        # with path
        ("domain.com/test", "http://domain.com/test"),
        # query
        ("domain.com/test?123", "http://domain.com/test?123"),
        # fragment
        ("domain.com/test#123", "http://domain.com/test#123"),
        # normalize domain, subdomain and scheme
        ("HTTP://SUB.DOMAIN.COM/TeSt", "http://sub.domain.com/TeSt"),
        # non-ascii and space in path
        ("domain.com/El Niño", "http://domain.com/El%20Ni%C3%B1o"),
        # non-ascii in domain and subdomain
        ("fóò.café.fr/test", "http://xn--f-tgac.xn--caf-dma.fr/test"),
        # with port
        ("a.com:1234", "http://a.com:1234/"),
        # port and path
        ("a.com:1234/c", "http://a.com:1234/c"),
        # port, path, parameters, query and fragment
        ("a.com:80/c;p?q=1&q2#f", "http://a.com:80/c;p?q=1&q2#f"),
    ],
)
def test_url_parse(url_str, expected):
    """test URL.parse()"""
    assert str(URL.parse(url_str)) == expected


@mark.parametrize(
    "url_str, msg",
    [
        ("", "Missing domain"),
        ("/", "Missing domain"),
        ("http://", "Missing domain"),
        ("user:name@", "Missing domain"),
        (".com", "Missing domain"),
        ("data:text/html,TEST", "Missing domain"),
        (".a.com", r"Invalid subdomain$"),
        ("test", "Missing suffix"),
        ("192.168.1.1", "Missing suffix"),
        ("about:config", "Missing suffix"),
        ("foo.notarealtld", "Missing suffix"),
        ("http://a.com:foo", "Port must be a number"),
        ("http://a.com:0", "Invalid port number"),
        ("http://[a.com/", "Invalid URL"),
    ],
)
def test_url_parse_invalid_strings(url_str, msg):
    """test URL.parse() - invalid strings()"""
    with raises(URLParseError, match=msg):
        URL.parse(url_str)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from argparse import ArgumentParser, Namespace
from importlib.metadata import PackageNotFoundError, version
from logging import DEBUG, INFO
from os import getenv
from pathlib import Path
from platform import system
from shutil import which

try:
    __version__ = version("site-scout")
except PackageNotFoundError:  # pragma: no cover
    # package is not installed
    __version__ = "unknown"

TIME_LIMIT_DEBUG = 300
TIME_LIMIT_DEFAULT = 90
TIME_LIMIT_EXPLORE = 180


def is_headless() -> bool:
    """Detect if running in a headless environment.

    Args:
        None

    Returns:
        True if running in a headless environment otherwise False.
    """
    return (
        system() == "Linux" and not getenv("DISPLAY") and not getenv("WAYLAND_DISPLAY")
    )


def parse_args(argv: list[str] | None = None) -> Namespace:
    """Argument parsing"""
    parser = ArgumentParser(
        prog="site-scout", description="Visit provided URLs and report results."
    )
    parser.add_argument("binary", type=Path, help="Firefox binary to use.")

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "-i",
        "--input",
        default=[],
        nargs="+",
        type=Path,
        help="File(s) containing URLs to visit.",
    )
    group.add_argument("-u", "--url", default=[], nargs="+", help="URL(s) to visit.")

    parser.add_argument(
        "-b",
        "--browser",
        choices=("firefox", "fenix"),
        default="firefox",
        help="Browser wrapper to use (autodetect is best effort).",
    )
    parser.add_argument(
        "--disable-logging",
        action="store_true",
        help="Disable console output before running (default: %(default)s).",
    )
    display_choices = ["default", "headless"]
    if system() == "Linux":
        display_choices.append("xvfb")
    parser.add_argument(
        "--display",
        choices=display_choices,
        default="xvfb" if is_headless() else "default",
        help="Display mode.",
    )
    parser.add_argument(
        "--explore",
        const="all",
        nargs="?",
        default=None,
        choices=("all", "load"),
        help="Use PageExplorer to interact with content. "
        "Options: 'all' run all operations, 'load' wait for page load and exit.",
    )
    parser.add_argument(
        "--fuzzmanager", action="store_true", help="Report results to FuzzManager."
    )
    parser.add_argument(
        "--jobs", type=int, default=1, help="Maximum number of browsers to run at once."
    )
    parser.add_argument(
        "--launch-timeout",
        type=int,
        default=300,
        help="Time in seconds to wait for browser to launch and begin navigation"
        " (default: %(default)s).",
    )
    level_map = {"INFO": INFO, "DEBUG": DEBUG}
    parser.add_argument(
        "--log-level",
        choices=sorted(level_map),
        default="INFO",
        help="Configure console output (default: %(default)s).",
    )
    parser.add_argument(
        "--memory-limit",
        type=int,
        default=0,
        help="Browser memory limit in MBs (default: no limit).",
    )
    parser.add_argument(
        "--omit-urls",
        action="store_true",
        help="Do not include URLs in saved results (default: %(default)s).",
    )
    parser.add_argument(
        "-o",
        "--output-path",
        default=Path.cwd(),
        type=Path,
        help="Location to save results (default: %(default)s).",
    )
    parser.add_argument(
        "--prefs",
        type=Path,
        help="Custom prefs.js file to use (default: generated).",
    )
    parser.add_argument("--profile", type=Path)
    parser.add_argument(
        "--result-limit",
        type=int,
        default=0,
        help="Maximum number of results."
        " Remaining visits will be skipped if limit is hit (default: no limit).",
    )
    parser.add_argument(
        "--runtime-limit",
        type=int,
        default=0,
        help="Maximum runtime in seconds."
        " Remaining visits will be skipped if limit is hit (default: no limit).",
    )
    parser.add_argument(
        "--status-report",
        type=Path,
        help="Location to save periodic status report (default: no report).",
    )
    parser.add_argument(
        "--time-limit",
        type=int,
        help=f"Page load time limit in seconds (default: {TIME_LIMIT_DEFAULT}).",
    )
    parser.add_argument(
        "--url-limit",
        type=int,
        default=0,
        help="Maximum number of URLs to visit (default: no limit).",
    )
    parser.add_argument(
        "--version",
        "-V",
        action="version",
        version=f"%(prog)s {__version__}",
        help="Show version number.",
    )
    parser.add_argument(
        "--visits",
        type=int,
        default=1,
        help="Number of times to visit each URL (default: %(default)s).",
    )

    parser.set_defaults(coverage=False, debugger=None)
    if system() == "Linux":
        parser.add_argument(
            "--coverage",
            action="store_true",
            help="Dump coverage data to disk. This requires a supported browser build.",
        )
        dbg_group = parser.add_mutually_exclusive_group()
        dbg_group.add_argument(
            "--pernosco",
            action="store_const",
            const="PERNOSCO",
            dest="debugger",
            help="Use rr. Trace intended to be used with Pernosco.",
        )
        dbg_group.add_argument(
            "--rr",
            action="store_const",
            const="RR",
            dest="debugger",
            help="Use rr.",
        )

    args = parser.parse_args(argv)

    if args.time_limit is None:
        if args.debugger is not None:
            args.time_limit = TIME_LIMIT_DEBUG
        elif args.explore:
            args.time_limit = TIME_LIMIT_EXPLORE
        else:
            args.time_limit = TIME_LIMIT_DEFAULT
    if args.time_limit < 1:
        parser.error("--time-limit must be > 0 (recommended minimum: 30)")

    if args.coverage:
        # GCOV_PREFIX_STRIP and GCOV_PREFIX are specific to Firefox coverage builds
        if not getenv("GCOV_PREFIX_STRIP"):
            parser.error("GCOV_PREFIX_STRIP must be set to use --coverage")
        if not getenv("GCOV_PREFIX"):
            parser.error("GCOV_PREFIX must be set to use --coverage")
        if args.jobs > 1:
            parser.error("Parallel jobs not supported with --coverage")

    if args.debugger in ("PERNOSCO", "RR"):
        if args.fuzzmanager:
            parser.error("rr not supported with --fuzzmanager")
        # rr is only supported on Linux
        if not which("rr"):
            parser.error("rr is not installed")
        settings = "/proc/sys/kernel/perf_event_paranoid"
        value = int(Path(settings).read_bytes())
        if value > 1:
            parser.error(f"rr needs {settings} <= 1, but it is {value}")

    if args.jobs < 1:
        parser.error("--jobs must be >= 1")

    args.log_level = level_map[args.log_level]

    if args.memory_limit < 0:
        parser.error("--memory-limit must be >= 0")
    args.memory_limit *= 1_048_576

    if not args.binary.is_file():
        parser.error(f"binary does not exist: '{args.binary}'")

    if args.binary.suffix.lower() == ".apk":
        args.browser = "fenix"

    if args.explore is not None and args.browser == "fenix":
        parser.error("--explore not supported with selected browser")

    for in_path in args.input:
        if not in_path.exists():
            parser.error(f"-i/--input does not exist: '{in_path}'")

    if not args.output_path.is_dir():
        parser.error(f"-o/--output-path does not exist: '{args.output_path}'")

    if args.prefs and not args.prefs.is_file():
        parser.error(f"--prefs does not exist: '{args.prefs}'")

    return args



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from argparse import ArgumentParser, Namespace
from bisect import insort
from logging import DEBUG, INFO, getLogger
from pathlib import Path
from typing import TYPE_CHECKING, cast

from yaml import dump

try:
    from yaml import CSafeDumper as SafeDumper
except ImportError:
    from yaml import SafeDumper  # type: ignore

from .main import init_logging, load_yml
from .url import URL, URLParseError
from .url_db import UrlDBError

if TYPE_CHECKING:
    from collections.abc import Generator

    from .url_db import UrlDB

LOG = getLogger(__name__)


class UrlCollection:
    """Manage collections of URLs."""

    __slots__ = ("_db", "unparsable")

    def __init__(self, url_db: UrlDB | None = None) -> None:
        self._db = url_db or cast("UrlDB", {})
        self.unparsable: set[str] = set()

    def __len__(self) -> int:
        count = 0
        for domain_entries in self._db.values():
            count += sum(len(paths) for paths in domain_entries.values())
        return count

    def __iter__(self) -> Generator[URL]:
        for domain, domain_entries in self._db.items():
            for subdomain, paths in domain_entries.items():
                for path in paths:
                    yield URL(domain, subdomain=subdomain, path=path)

    def add_list(self, urls_file: Path) -> int:
        """Load a text file containing a line separated list of URLs and add them
        to the collection of URLs.

        Args:
            urls_file: File to load.

        Returns:
            Number of previously unknown URLs that have been added.
        """
        added = 0
        with urls_file.open() as in_fp:
            for line in in_fp:
                entry = line.strip()
                if not entry or entry[0] == "#":
                    # skip comments and empty lines
                    continue
                if self.add_str(entry) is not None:
                    added += 1
        return added

    def add_str(self, url: str) -> URL | None:
        """Parse URL from a given string and add it to the collection of known URLs.

        Args:
            url: URL to add.

        Returns:
            URL that was added.
        """
        try:
            parsed = URL.parse(url)
        except URLParseError:
            LOG.debug("failed to parse and add: '%s'", url)
            self.unparsable.add(url)
            return None
        return parsed if self.add_url(parsed) else None

    def add_url(self, url: URL) -> bool:
        """Add a URL to the collection of known URLs.

        Args:
            url: URL to add.

        Returns:
            True if the URL was added otherwise False.
        """
        if url.domain not in self._db:
            self._db[url.domain] = {}
        if url.subdomain not in self._db[url.domain]:
            self._db[url.domain][url.subdomain] = []
        if url.path not in self._db[url.domain][url.subdomain]:
            insort(self._db[url.domain][url.subdomain], url.path)
            LOG.debug("added: %s", url)
            return True
        return False

    def deduplicate_www(self) -> int:
        """Scan domain entries with no subdomain and 'www' and remove duplicate entries
        favoring 'www'. For example if both 'mozilla.org/' and 'www.mozilla.org/' exist
        only 'www.mozilla.org/' will be kept.

        Args:
            None

        Returns:
            Total URLs deduplicated.
        """
        count = 0
        # iterate over all domains
        for name, entries in self._db.items():
            if "www" in entries and "" in entries:
                # iterate over duplicate paths and remove them
                for path in tuple(x for x in entries[""] if x in entries["www"]):
                    count += 1
                    entries[""].remove(path)
                    LOG.debug("deduped www.%s%s", name, path)
                # remove empty subdomain entry if needed
                if not entries[""]:
                    del entries[""]
        return count

    @property
    def domains(self) -> Generator[str]:
        """All known domains.

        Args:
            None

        Yields:
            Domains in URL collection.
        """
        yield from self._db

    def count_entries(self) -> dict[str, int]:
        """Count entries for each domain.

        Args:
            None

        Returns:
            A mapping of domains and entry counts.
        """
        results: dict[str, int] = {}
        for domain, domain_entries in self._db.items():
            results[domain] = sum(len(paths) for paths in domain_entries.values())
        return results

    def remove_url(self, url: str) -> bool:
        """Remove a URL from the collection.

        Args:
            url: URL to remove.

        Returns:
            True if the URL was removed otherwise False.
        """
        try:
            parsed = URL.parse(url)
        except URLParseError:
            LOG.debug("failed to parse and remove: '%s'", url)
            return False

        try:
            self._db[parsed.domain][parsed.subdomain].remove(parsed.path)
        except (KeyError, ValueError):
            LOG.debug("failed to remove unknown url: '%s'", url)
            return False

        # remove empty domain/subdomain entries
        if not self._db[parsed.domain][parsed.subdomain]:
            del self._db[parsed.domain][parsed.subdomain]
        if not self._db[parsed.domain]:
            del self._db[parsed.domain]

        return True

    @classmethod
    def load_yml(cls, src: Path) -> UrlCollection | None:
        """Load UrlCollection from a YML file.

        Args:
            src: File to load.

        Returns:
            UrlCollection.
        """
        try:
            return cls(load_yml(src))
        except UrlDBError as exc:
            LOG.warning("Load failure - %s (ignored: %s)", exc, src.resolve())
        return None

    def save_yml(self, dst: Path) -> None:
        """Save UrlCollection to a YML file.

        Args:
            dst: File to save to.

        Returns:
            None
        """
        with dst.open("w") as out_fp:
            dump(self._db, out_fp, Dumper=SafeDumper, default_style="'")

    def sort_paths(self) -> None:
        """Sort paths.

        Args:
            None

        Returns:
            None.
        """
        for domain_entries in self._db.values():
            for paths in domain_entries.values():
                paths.sort()


def parse_args(argv: list[str] | None = None) -> Namespace:
    """Argument parsing"""
    parser = ArgumentParser(description="Manage URL collections used by Site-scout.")
    parser.add_argument("url_db", type=Path, help="YML file containing URL data.")
    parser.add_argument(
        "--disable-logging",
        action="store_true",
        help="Disable console output before running (default: %(default)s).",
    )
    parser.add_argument("-d", "--display-list", action="store_true")
    parser.add_argument(
        "--domain-entries",
        type=int,
        default=0,
        help="Display domains with the most entries (default: %(default)s).",
    )
    parser.add_argument(
        "--force-save",
        action="store_true",
        help="Save data even if no changes have been made (refresh YML file).",
    )
    level_map = {"INFO": INFO, "DEBUG": DEBUG}
    parser.add_argument(
        "--log-level",
        choices=sorted(level_map),
        default="INFO",
        help="Configure console logging (default: %(default)s).",
    )
    group = parser.add_mutually_exclusive_group(required=False)
    group.add_argument(
        "-l",
        "--url-list",
        type=Path,
        help="A .yml file URL database or a .txt file containing line separated list"
        "of URLs.",
    )
    group.add_argument("-r", "--remove", help="Remove single URL.")
    group.add_argument("-u", "--url", help="Add single URL.")
    args = parser.parse_args(argv)

    args.log_level = level_map[args.log_level]

    if args.url_list and not args.url_list.is_file():
        parser.error(f"{args.url_list} is not a file.")

    return args


def main(argv: list[str] | None = None) -> int:
    """Main function"""
    args = parse_args(argv)
    init_logging(args.log_level, args.disable_logging)

    # pylint: disable=too-many-nested-blocks
    try:
        # load data
        if args.url_db and args.url_db.is_file():
            LOG.info("Loading '%s'...", args.url_db)
            urls = UrlCollection.load_yml(args.url_db)
            if urls is None:
                return 1
            LOG.debug("sorting paths...")
            urls.sort_paths()
        else:
            urls = UrlCollection()

        # update data
        changed = False
        if args.url:
            added = urls.add_str(args.url)
            if added is not None:
                LOG.info("Added '%s'", added)
                changed = True
        elif args.url_list:
            LOG.info("Scanning URL(s) in '%s'...", args.url_list)
            if args.url_list.suffix.lower() == ".yml":
                # attempt to load URLs from a YML file
                merge_urls = UrlCollection.load_yml(args.url_list)
                if merge_urls is not None:
                    added_count = 0
                    for entry in merge_urls:
                        if urls.add_url(entry):
                            added_count += 1
            else:
                # attempt to load a line separated list of URLs
                added_count = urls.add_list(args.url_list)
            LOG.info("Added %s URL(s).", f"{added_count:,d}")
            changed = added_count > 0
        elif args.remove:
            changed = urls.remove_url(args.remove)
            if changed:
                LOG.info("Removed 1 URL.")

        LOG.debug("deduplicating entries...")
        deduped = urls.deduplicate_www()
        if deduped > 0:
            LOG.info("Total URLs deduplicated: %s.", f"{deduped:,d}")
            changed = True

        # save data
        if changed or args.force_save:
            LOG.info("Saving '%s'...", args.url_db)
            urls.save_yml(args.url_db)
        elif not args.url_db.is_file():
            LOG.error("File not found '%s'", args.url_db)
            return 1
        else:
            LOG.info("No change made.")

        # show current list of URLs
        if args.display_list:
            LOG.info("Known URLs:")
            for entry in urls:
                LOG.info("> %s", entry)

        # show unparsable input
        if urls.unparsable:
            LOG.warning("Unparsable entries: %s", f"{len(urls.unparsable):,d}")
            for unparsable in urls.unparsable:
                LOG.info("> '%s'", unparsable)

        # largest domains
        if args.domain_entries > 0:
            LOG.info("Domains with most entries:")
            for domain, count in sorted(
                urls.count_entries().items(), key=lambda kv: kv[1], reverse=True
            )[: args.domain_entries]:
                LOG.info("> %d '%s'", count, domain)

        # show summary
        file_size = args.url_db.stat().st_size
        LOG.info(
            "%s URLs (%s domains) in db (%s bytes).",
            f"{len(urls):,d}",
            f"{sum(1 for _ in urls.domains):,d}",
            f"{file_size:,d}",
        )

    except KeyboardInterrupt:  # pragma: no cover
        LOG.warning("Aborting...")

    return 0


if __name__ == "__main__":
    raise SystemExit(main())



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from logging import getLogger
from os import getenv
from shutil import copyfile
from subprocess import TimeoutExpired
from time import perf_counter
from typing import TYPE_CHECKING

from fxpoppet.adb_process import ADBLaunchError, ADBProcess, Reason
from fxpoppet.adb_session import ADBSession, ADBSessionError
from fxpoppet.emulator.android import AndroidEmulator, AndroidEmulatorError

from .browser_wrapper import (
    BrowserArgs,
    BrowserState,
    BrowserWrapper,
    EnvironmentManager,
)

if TYPE_CHECKING:
    from pathlib import Path

    from .explorer import Explorer, ExplorerMode
    from .url import URL


LOG = getLogger(__name__)


class FenixWrapper(BrowserWrapper):
    """FenixWrapper is used to add support for Firefox for Android."""

    __slots__ = ("_device", "_env_mgr", "_package", "_proc", "_session")

    def __init__(
        self, args: BrowserArgs, working_path: Path, env_mgr: FenixEnvironmentManager
    ) -> None:
        super().__init__(args, working_path, env_mgr)
        self._env_mgr = env_mgr
        package = ADBSession.get_package_name(args.binary)
        if package is None:
            LOG.error("FenixWrapper failed to find package name!")
            raise RuntimeError("Could not find package name.")
        self._package = package
        self._proc: ADBProcess | None = None
        self._session: ADBSession | None = None
        # TODO: generate prefs if needed
        self._device: str | None = None

    def cleanup(self) -> None:
        if self._proc is not None:
            self._proc.cleanup()
        if self._session is not None:
            self._session.reverse_remove()
        if self._device is not None:
            self._env_mgr.release_device(self._device)

    def close(self) -> None:
        if self._proc is not None:
            self._proc.close()

    def create_explorer(
        self, url: URL, load_wait: int, mode: ExplorerMode, pause: int = 0
    ) -> Explorer:
        raise NotImplementedError()

    def debugger(self) -> bool:
        return False

    def dump_coverage(self, timeout: int = 15) -> None:
        raise NotImplementedError()

    @staticmethod
    def environment_manager(
        instance_limit: int, browser_args: BrowserArgs
    ) -> FenixEnvironmentManager:
        return FenixEnvironmentManager(instance_limit, browser_args)

    def is_healthy(self) -> bool:
        assert self._proc is not None
        return self._proc.is_healthy()

    def is_idle(self, idle_limit: int) -> bool:
        assert self._proc is not None
        if self._proc.is_healthy():
            # check that no processes are over the limit
            return not any(pct > idle_limit for _, pct in self._proc.cpu_usage())
        # not running
        return True

    def launch(
        self,
        url: str,
        explore: bool,
        log_path: Path | None,
        raise_on_failure: bool,
    ) -> bool:
        assert not explore, "explore not supported"
        assert self._proc is None
        assert self._session is None
        # select a device
        self._device = self._env_mgr.select_device(self.args.binary)
        if self._device is None:
            LOG.error("Failed to select target device!")
            return False
        # connect to the device
        try:
            session = ADBSession.connect(self._device, as_root=True, boot_timeout=10)
        except ADBSessionError as exc:
            LOG.error("FenixWrapper failed to connect to device: %s", exc)
            self._env_mgr.shutdown_device(self._device)
            return False
        LOG.debug("connected to '%s'", self._device)
        self._session = session
        self._session.symbols[self._package] = self.args.binary.parent / "symbols"
        try:
            self._proc = ADBProcess(self._package, self._session)
        except ADBSessionError as exc:
            LOG.error("FenixWrapper launch failed: %s", exc)
            self._env_mgr.shutdown_device(self._device)
            return False
        # launch the browser on the device
        try:
            self._proc.launch(
                str(url),
                env_mod={
                    # prevent crash reporter from touching the dmp files
                    "MOZ_CRASHREPORTER": "1",
                    "MOZ_CRASHREPORTER_NO_REPORT": "1",
                    "MOZ_CRASHREPORTER_SHUTDOWN": "1",
                },
                launch_timeout=self.args.launch_timeout,
                prefs_js=self.args.prefs_file,
            )
        except ADBLaunchError as exc:
            LOG.error("ADBLaunchError: %s", exc)
            self.close()
            self._env_mgr.shutdown_device(self._device)
            if raise_on_failure:
                raise
            return False
        return True

    def save_report(self, dst: Path) -> None:
        assert self._proc is not None
        self._proc.save_logs(dst)
        # save prefs file
        if self.args.prefs_file is not None:
            copyfile(self.args.prefs_file, dst / "prefs.js")

    def state(self) -> BrowserState:
        assert self._proc is not None
        assert self._proc.reason is not None
        if self._proc.reason == Reason.EXITED:
            return BrowserState.EXITED
        if self._proc.reason == Reason.ALERT:
            return BrowserState.RESULT
        return BrowserState.CLOSED

    def wait(self, timeout: float) -> bool:
        assert self._proc is not None
        return self._proc.wait(timeout=timeout)


class EmulatorPool:
    """Create and manage a collection of Android emulators.

    Attributes:
        _display_mode: Display mode of emulators.
        _emulators: Emulator objects.
        _in_use: Corresponding serials of emulators currently in use.
        _launch_failures: Current consecutive launch failure count.
        _size_limit: Maximum number of emulators allowed.
    """

    # consecutive launch failure limit
    LAUNCH_FAILURE_LIMIT = 3

    __slot__ = (
        "_display_mode",
        "_emulators",
        "_in_use",
        "_launch_failures",
        "_size_limit",
    )

    def __init__(self, size_limit: int, display_mode: str = "default") -> None:
        assert display_mode in {"default", "headless", "xvfb"}
        assert size_limit > 0
        self._display_mode = display_mode
        self._emulators: dict[str, AndroidEmulator] = {}
        self._in_use: set[str] = set()
        self._launch_failures = 0
        self._size_limit = size_limit

    def cleanup(self) -> None:
        """Terminate all running emulators.

        Args:
            None

        Returns:
            None
        """
        for emu in self._emulators.values():
            if emu.poll() is None:
                emu.terminate()
        deadline = perf_counter() + 60
        for emu in self._emulators.values():
            max_delay = max(deadline - perf_counter(), 1)
            try:
                emu.wait(timeout=max_delay)
            except TimeoutExpired:
                LOG.warning("Failed to terminate emulator")
                emu.emu.kill()
                continue
            emu.cleanup()

    def manage_emulators(self, apk: Path) -> None:
        """Manage the emulator pool.

        Args:
            apk: Browser APK file.

        Returns:
            None
        """
        # check emulators are in a good state
        self._check_emulators()
        # remove emulators that are not running from the pool
        self._trim_emulators()
        # launch emulator if needed
        available = len(self._emulators) - len(self._in_use)
        assert available >= 0
        LOG.debug(
            "emulators: running %d, available %d, limit %d",
            len(self._emulators),
            available,
            self._size_limit,
        )
        if available == 0 and len(self._emulators) < self._size_limit:
            try:
                emu = self._launch_emulator()
            except AndroidEmulatorError as exc:
                LOG.warning("Failed to launch emulator: %s", exc)
                self._launch_failures += 1
                if self._launch_failures >= self.LAUNCH_FAILURE_LIMIT:
                    raise
                return
            self._launch_failures = 0
            serial = f"emulator-{emu.port}"
            if not self._prepare_emulator(emu, apk):
                self._shutdown_emulator(emu)
                LOG.warning("Failed to prepare device '%s'", serial)
                return
            assert serial not in self._emulators
            self._emulators[serial] = emu
            assert len(self._emulators) <= self._size_limit

    def release(self, serial: str) -> None:
        """Mark an emulator as not in use.

        Args:
            serial: Emulator to mark as not in use.

        Returns:
            None
        """
        assert serial in self._in_use
        self._in_use.remove(serial)

    def select(self) -> str | None:
        """Select available emulator.

        Args:
            None.

        Returns:
            None
        """
        available = set(self._emulators) - self._in_use
        while available:
            selected = available.pop()
            # check if emulator is running
            if self._emulators[selected].poll() is None:
                self._in_use.add(selected)
                return selected
        LOG.debug("no running emulator available")
        return None

    def shutdown(self, serial: str) -> None:
        """Shutdown an emulator.

        Args:
            serial: Emulator to shutdown.

        Returns:
            None
        """
        self._shutdown_emulator(self._emulators[serial])

    def _check_emulators(self) -> None:
        for serial, emu in self._emulators.items():
            if emu.poll() is not None:
                # emulator is not running
                continue
            try:
                ADBSession.connect(serial, as_root=False, boot_timeout=10)
            except ADBSessionError as exc:
                LOG.warning("Cannot connect to emulator '%s': %s", serial, exc)
                self._shutdown_emulator(emu)

    def _launch_emulator(self) -> AndroidEmulator:
        port = AndroidEmulator.search_free_ports()
        avd_name = f"x86.{port:d}"
        LOG.debug("launching 'emulator-%d'...", port)
        AndroidEmulator.create_avd(avd_name)
        try:
            emu = AndroidEmulator(
                port=port,
                avd_name=avd_name,
                boot_timeout=300,
                headless=self._display_mode == "headless",
                xvfb=self._display_mode == "xvfb",
                emulator_output=getenv("EMULATOR_OUTPUT", "") == "1",
            )
        except:
            AndroidEmulator.remove_avd(avd_name)
            raise
        return emu

    def _prepare_emulator(self, emu: AndroidEmulator, apk: Path) -> bool:
        serial = f"emulator-{emu.port}"
        LOG.debug("preparing '%s'...", serial)
        try:
            session = ADBSession.connect(serial, as_root=True, boot_timeout=10)
        except ADBSessionError as exc:
            LOG.debug("cannot connect to prepare emulator '%s': %s", serial, exc)
            return False
        # disable some UI animations
        session.device.shell(
            ["settings", "put", "global", "animator_duration_scale", "0"]
        )
        session.device.shell(
            ["settings", "put", "global", "transition_animation_scale", "0"]
        )
        session.device.shell(
            ["settings", "put", "global", "window_animation_scale", "0"]
        )
        # prevent device throttling
        session.device.shell(["settings", "put", "global", "device_idle_enabled", "0"])
        session.device.shell(["settings", "put", "global", "low_power", "0"])
        session.device.shell(
            ["settings", "put", "global", "background_process_limit", "0"]
        )
        session.device.shell(["dumpsys", "deviceidle", "disable"])
        try:
            session.install(apk)
        except ADBSessionError as exc:
            LOG.debug("failed to install APK: %s", exc)
            return False
        return True

    @staticmethod
    def _shutdown_emulator(emu: AndroidEmulator) -> None:
        if emu.poll() is None:
            emu.terminate()
        try:
            emu.wait(timeout=30)
        except TimeoutExpired:
            emu.emu.kill()
            try:
                emu.wait(timeout=30)
            except TimeoutExpired:
                LOG.warning("Failed to shutdown emulator")
        finally:
            emu.cleanup()

    def _trim_emulators(self) -> None:
        # scan pool for emulators that are not running
        not_running: list[str] = []
        for serial, emu in self._emulators.items():
            exit_code = emu.poll()
            if exit_code is not None:
                LOG.debug("removing pool entry: '%s' (%d)", serial, exit_code)
                not_running.append(serial)
        # remove emulators that are not running from the pool
        for serial in not_running:
            if serial in self._in_use:
                LOG.warning("Emulator '%s' not running and in use!", serial)
                self._in_use.remove(serial)
            self._emulators.pop(serial)


class FenixEnvironmentManager(EnvironmentManager):
    """FenixEnvironmentManager is used to add support for Firefox for Android."""

    def __init__(self, instance_limit: int, browser_args: BrowserArgs) -> None:
        super().__init__()
        self._pool = EmulatorPool(
            instance_limit, display_mode=browser_args.display_mode or "default"
        )

    def select_device(self, apk: Path) -> str | None:
        """Select available device.

        Args:
            apk: Browser APK file.

        Returns:
            Device serial if a device is available otherwise None.
        """
        self._pool.manage_emulators(apk)
        return self._pool.select()

    def release_device(self, serial: str) -> None:
        """Release device.

        Args:
            serial: Device to mark as available.

        Returns:
            None
        """
        self._pool.release(serial)

    def shutdown_device(self, serial: str) -> None:
        """Shutdown device.

        Args:
            serial: Device to shutdown.

        Returns:
            None
        """
        self._pool.shutdown(serial)

    def cleanup(self) -> None:
        self._pool.cleanup()



# type: ignore
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
from subprocess import TimeoutExpired

from fxpoppet import ADBLaunchError, ADBProcess, Reason
from fxpoppet.adb_session import ADBSessionError
from fxpoppet.emulator.android import AndroidEmulator, AndroidEmulatorError
from pytest import mark, raises

from .browser_wrapper import BrowserArgs, BrowserState
from .fenix_wrapper import EmulatorPool, FenixEnvironmentManager, FenixWrapper
from .url import URL


def test_fenix_wrapper_invalid_apk(mocker, tmp_path):
    """test FenixWrapper invalid APK"""
    session = mocker.patch("site_scout.fenix_wrapper.ADBSession", autospec=True)
    session.get_package_name.return_value = None
    with raises(RuntimeError, match=r"Could not find package name."):
        FenixWrapper(
            BrowserArgs(tmp_path / "target.apk", 10, 10),
            tmp_path,
            mocker.Mock(spec_set=FenixEnvironmentManager),
        )


def test_fenix_wrapper_basic(mocker, tmp_path):
    """test FenixWrapper basic"""
    prefs = tmp_path / "custom_prefs.js"
    prefs.touch()
    mocker.patch("site_scout.fenix_wrapper.ADBSession", autospec=True)
    proc = mocker.patch(
        "site_scout.fenix_wrapper.ADBProcess", autospec=True
    ).return_value
    with FenixWrapper(
        BrowserArgs(tmp_path / "target.apk", 10, 10, prefs_file=prefs),
        tmp_path,
        mocker.Mock(spec_set=FenixEnvironmentManager),
    ) as browser:
        assert not browser.debugger()
        # launch
        browser.launch(URL("a.com"), False, None, False)
        assert proc.launch.call_count == 1
        assert proc.cleanup.call_count == 0
        # is_healthy
        proc.is_healthy.return_value = True
        assert browser.is_healthy()
        assert proc.is_healthy.call_count == 1
        # save report
        browser.save_report(tmp_path)
        assert (tmp_path / "prefs.js").is_file()
        # wait wrapper test
        browser.wait(0)
        assert proc.wait.call_count == 1
        # is_idle should check cpu usage when is_healthy is true
        proc.cpu_usage.return_value = ((123, 99),)
        assert not browser.is_idle(idle_limit=10)
        assert proc.cpu_usage.call_count == 1
        proc.is_healthy.return_value = False
        assert browser.is_idle(idle_limit=10)
        # close
        browser.close()
        assert proc.close.call_count == 1
    assert proc.cleanup.call_count == 1


@mark.parametrize(
    "state, reason",
    [
        (BrowserState.EXITED, Reason.EXITED),
        (BrowserState.RESULT, Reason.ALERT),
        (BrowserState.CLOSED, Reason.CLOSED),
    ],
)
def test_fenix_wrapper_state(mocker, tmp_path, state, reason):
    """test FenixWrapper.state()"""
    mocker.patch("site_scout.fenix_wrapper.ADBSession", autospec=True)
    mocker.patch("site_scout.fenix_wrapper.FenixEnvironmentManager", autospec=True)
    args = BrowserArgs(tmp_path / "target.apk", 10, 10)
    env_mgr = FenixWrapper.environment_manager(1, args)
    with FenixWrapper(args, tmp_path, env_mgr) as browser:
        browser._proc = mocker.Mock(spec_set=ADBProcess)
        browser._proc.reason = reason
        assert browser.state() == state


def test_fenix_wrapper_launch_failures(mocker, tmp_path):
    """test FenixWrapper.launch() failures"""
    session_cls = mocker.patch("site_scout.fenix_wrapper.ADBSession", autospec=True)
    proc_cls = mocker.patch("site_scout.fenix_wrapper.ADBProcess", autospec=True)
    env_mgr = mocker.Mock(spec_set=FenixEnvironmentManager)
    with FenixWrapper(
        BrowserArgs(tmp_path / "target.apk", 10, 10), tmp_path, env_mgr
    ) as browser:
        # no device
        env_mgr.select_device.return_value = None
        assert not browser.launch(URL("a.com"), False, None, False)
        # failed to create session
        env_mgr.reset_mock(return_value=True)
        session_cls.connect.side_effect = ADBSessionError("boot failed")
        assert not browser.launch(URL("a.com"), False, None, False)
        assert proc_cls.call_count == 0
        session_cls.reset_mock(side_effect=True)
        # failed to create proc
        proc_cls.side_effect = ADBLaunchError("launch-fail-test")
        assert not browser.launch(URL("a.com"), False, None, False)
        assert browser._session is not None
        assert browser._proc is None
        assert proc_cls.call_count == 1
        proc_cls.reset_mock(side_effect=True)
        # failed to launch (raise false)
        proc_cls.return_value.launch.side_effect = ADBLaunchError("launch-fail-test")
        browser._session = None
        assert not browser.launch(URL("a.com"), False, None, False)
        assert browser._session is not None
        assert browser._proc is not None
        # failed to launch (raise true)
        browser._session = None
        browser._proc = None
        with raises(ADBLaunchError, match="launch-fail-test"):
            browser.launch(URL("a.com"), False, None, True)


def test_emulator_pool_launch_and_prep_emulator(mocker, tmp_path):
    """test EmulatorPool launch and prepare emulator"""
    emu_cls = mocker.patch("site_scout.fenix_wrapper.AndroidEmulator", autospec=True)
    emu_cls.search_free_ports.return_value = 1234
    session_cls = mocker.patch("site_scout.fenix_wrapper.ADBSession", autospec=True)
    pool = EmulatorPool(1)
    # handle a launch failure
    emu_cls.side_effect = RuntimeError("testing-failed-launch")
    with raises(RuntimeError, match="testing-failed-launch"):
        pool._launch_emulator()
    assert emu_cls.create_avd.call_count == 1
    assert emu_cls.remove_avd.call_count == 1
    # handle process shutdown
    session_cls.connect.side_effect = ADBSessionError("boot failed")
    emu_cls.reset_mock(side_effect=True)
    emu = pool._launch_emulator()
    assert emu_cls.create_avd.call_count == 1
    assert emu_cls.remove_avd.call_count == 0
    assert not pool._prepare_emulator(emu, tmp_path / "fake.apk")
    assert session_cls.connect.call_count == 1
    assert session_cls.connect.return_value.device.shell.call_count == 0
    # success
    session_cls.reset_mock(side_effect=True)
    assert pool._prepare_emulator(emu, tmp_path / "fake.apk")
    assert session_cls.connect.call_count == 1
    assert session_cls.connect.return_value.device.shell.call_count == 7
    assert session_cls.connect.return_value.install.call_count == 1
    # prep failure
    session_cls.reset_mock()
    session_cls.connect.return_value.install.side_effect = ADBSessionError(
        "install failed"
    )
    assert not pool._prepare_emulator(emu, tmp_path / "fake.apk")
    assert session_cls.connect.call_count == 1


def test_emulator_pool_basic(mocker, tmp_path):
    """test EmulatorPool basic"""
    pool = EmulatorPool(1)
    assert pool._size_limit == 1
    assert not pool._in_use
    # launch failure
    mocker.patch.object(
        EmulatorPool, "_launch_emulator", side_effect=AndroidEmulatorError()
    )
    for _ in range(EmulatorPool.LAUNCH_FAILURE_LIMIT - 1):
        pool.manage_emulators(tmp_path / "target.apk")
    with raises(AndroidEmulatorError):
        pool.manage_emulators(tmp_path / "target.apk")
    # prepare fails
    emu = mocker.Mock(spec_set=AndroidEmulator)
    mocker.patch.object(EmulatorPool, "_launch_emulator", return_value=emu)
    mocker.patch.object(EmulatorPool, "_prepare_emulator", return_value=False)
    emu.poll.return_value = None
    pool.manage_emulators(tmp_path / "target.apk")
    assert emu.terminate.call_count == 1
    assert not pool._emulators
    # prepare success
    mocker.patch.object(EmulatorPool, "_prepare_emulator", return_value=True)
    emu.reset_mock(return_value=True)
    emu.port = 1234
    emu.poll.return_value = None
    pool.manage_emulators(tmp_path / "target.apk")
    assert pool._emulators
    assert "emulator-1234" in pool._emulators
    # select
    device = pool.select()
    assert device == "emulator-1234"
    assert "emulator-1234" in pool._in_use
    # release
    pool.release("emulator-1234")
    assert not pool._in_use
    # select - emulators not running
    pool._emulators["emulator-1234"].poll.return_value = 0
    assert pool.select() is None
    assert not pool._in_use
    # shutdown
    pool._emulators["emulator-1234"].poll.return_value = None
    emu.reset_mock()
    pool.shutdown("emulator-1234")
    assert emu.terminate.call_count == 1
    assert emu.wait.call_count == 1
    assert emu.cleanup.call_count == 1
    # cleanup
    emu.reset_mock()
    pool.cleanup()
    assert emu.terminate.call_count == 1
    assert emu.wait.call_count == 1
    assert emu.cleanup.call_count == 1
    # cleanup - failed to terminate
    emu.reset_mock()
    pool._emulators["emulator-1234"].poll.return_value = None
    pool._emulators["emulator-1234"].wait.side_effect = TimeoutExpired("foo", 1)
    pool.cleanup()
    assert emu.terminate.call_count == 1
    assert emu.wait.call_count == 1
    assert emu.emu.kill.call_count == 1


def test_emulator_pool_check_emulators(mocker):
    """test EmulatorPool._check_emulators()"""
    session_cls = mocker.patch("site_scout.fenix_wrapper.ADBSession", autospec=True)
    pool = EmulatorPool(1)
    # empty pool
    pool._check_emulators()
    # emulator not running
    emu = mocker.Mock(spec_set=AndroidEmulator)
    emu.poll.return_value = 0
    pool._emulators["test-1234"] = emu
    pool._check_emulators()
    assert session_cls.connect.call_count == 0
    # running emulator
    emu = mocker.Mock(spec_set=AndroidEmulator)
    emu.poll.return_value = None
    pool._emulators["test-1234"] = emu
    pool._check_emulators()
    assert session_cls.connect.call_count == 1
    assert emu.cleanup.call_count == 0
    # running emulator - connection failed
    session_cls.connect.side_effect = ADBSessionError("boot failed")
    pool._check_emulators()
    assert emu.cleanup.call_count == 1


def test_emulator_pool_trim_emulators(mocker):
    """test EmulatorPool._trim_emulators()"""
    session_cls = mocker.patch("site_scout.fenix_wrapper.ADBSession", autospec=True)
    pool = EmulatorPool(1)
    # empty pool
    pool._trim_emulators()
    # emulator running
    emu = mocker.Mock(spec_set=AndroidEmulator)
    emu.poll.return_value = None
    pool._emulators["test-1234"] = emu
    pool._trim_emulators()
    assert pool._emulators
    # emulator not running
    emu = mocker.Mock(spec_set=AndroidEmulator)
    emu.poll.return_value = 0
    pool._emulators["test-1234"] = emu
    pool._in_use.add("test-1234")
    pool._trim_emulators()
    assert not pool._emulators
    assert not pool._in_use
    assert session_cls.call_count == 0


def test_emulator_pool_shutdown_emulator(mocker):
    """test EmulatorPool._shutdown_emulator()"""
    # successful termination
    emu = mocker.Mock(spec_set=AndroidEmulator)
    emu.poll.return_value = None
    EmulatorPool._shutdown_emulator(emu)
    assert emu.terminate.call_count == 1
    assert emu.wait.call_count == 1
    assert emu.emu.kill.call_count == 0
    assert emu.cleanup.call_count == 1
    # failed termination
    emu.reset_mock()
    emu.wait.side_effect = TimeoutExpired("foo", 1)
    EmulatorPool._shutdown_emulator(emu)
    assert emu.terminate.call_count == 1
    assert emu.wait.call_count == 2
    assert emu.emu.kill.call_count == 1
    assert emu.cleanup.call_count == 1


def test_fenix_environment_manager_basic(mocker, tmp_path):
    """test FenixEnvironmentManager basic"""
    pool_cls = mocker.patch("site_scout.fenix_wrapper.EmulatorPool", autospec=True)
    pool_cls.return_value.select.return_value = "test-1234"
    env_mgr = FenixEnvironmentManager(1, BrowserArgs(tmp_path / "a.apk", 1, 1))
    assert env_mgr.select_device(tmp_path / "foo.apk") == "test-1234"
    env_mgr.release_device("test-1234")
    assert pool_cls.return_value.release.call_count == 1
    env_mgr.shutdown_device("test-1234")
    assert pool_cls.return_value.shutdown.call_count == 1
    env_mgr.cleanup()
    assert pool_cls.return_value.cleanup.call_count == 1



# type: ignore
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
from page_explorer import PageLoad
from pytest import mark, raises

from .explorer import Explorer, ExplorerError, ExplorerMode, State


@mark.parametrize(
    "get_value, explore_return, is_connected, state, title",
    (
        # navigation failed
        (ExplorerError(), False, [], State.LOADING, None),
        # failed during load
        (PageLoad.SUCCESS, False, [False], State.LOADING, "title"),
        # failed during load delay
        (PageLoad.SUCCESS, False, [True, False], State.LOADING, "title"),
        # failed to get page title
        (PageLoad.SUCCESS, True, [True] * 4 + [False], State.CLOSED, None),
        # server not found
        (PageLoad.FAILURE, False, [], State.NOT_FOUND, "Server Not Found"),
        # problem loading page
        (PageLoad.FAILURE, False, [], State.LOAD_FAILURE, "Problem loading page"),
        # unknown navigate/load failure
        (PageLoad.FAILURE, False, [], State.UNHANDLED_ERROR, "error title"),
        # failed during "skip to content"
        (PageLoad.SUCCESS, False, [True] * 2 + [False], State.SKIP_CONTENT, "title"),
        # browser crashed
        (PageLoad.SUCCESS, False, [True] * 3 + [False], State.EXPLORING, "title"),
        # successful interaction and browser crashed after explore
        (PageLoad.SUCCESS, True, [True] * 3 + [False], State.EXPLORING, "title"),
        # successful interaction and browser closed
        (PageLoad.SUCCESS, True, [True] * 4 + [False], State.CLOSED, "title"),
        # get timeout
        (PageLoad.TIMEOUT, True, [True] * 4 + [False], State.CLOSED, "title"),
    ),
)
def test_explorer(
    mocker, tmp_path, get_value, explore_return, is_connected, state, title
):
    """test Explorer()"""
    page_explorer = mocker.patch(
        "site_scout.explorer.PageExplorer", autospec=True
    ).return_value.__enter__.return_value
    page_explorer.current_url = "http://foo.foo"
    page_explorer.explore.return_value = explore_return
    page_explorer.get.side_effect = (get_value,)
    page_explorer.is_connected.side_effect = is_connected
    page_explorer.title = title
    with Explorer(tmp_path, 0, "http://foo.foo", pause=0) as explorer:
        assert not explorer._can_skip.is_set()
        # allow explore thread to complete
        explorer._thread.join(timeout=10)
        explorer.close()
        assert not explorer.is_running()
        assert explorer._can_skip.is_set()
        assert explorer.status.state == state
        if get_value in (PageLoad.SUCCESS, PageLoad.TIMEOUT):
            assert explorer.status.load_duration > 0
            assert explorer.status.url_loaded == "http://foo.foo"
        else:
            assert explorer.status.load_duration is None
            assert explorer.status.url_loaded is None
        if explore_return:
            assert explorer.status.explore_duration > 0
        else:
            assert explorer.status.explore_duration is None


def test_explorer_failed_create_page_explorer(mocker, tmp_path):
    """test Explorer() failed to create PageExplorer"""
    mocker.patch("site_scout.explorer.PageExplorer", side_effect=ExplorerError("test"))
    with Explorer(tmp_path, 0, "http://foo.foo") as explorer:
        with raises(RuntimeError, match=r"Explorer\.close\(\) not called"):
            assert explorer.status is None
        explorer.close()
        assert explorer.status.state == State.INITIALIZING
        assert explorer.status.load_duration is None
        assert explorer.status.explore_duration is None


def test_explorer_failed_init(mocker, tmp_path):
    """test Explorer() failed to create PageExplorer"""
    mocker.patch("site_scout.explorer.PageExplorer", autospec=True)
    fake_event = mocker.patch("site_scout.explorer.Event", autospec=True)
    fake_event.return_value.wait.return_value = False
    with (
        raises(RuntimeError, match="PageExplorer thread did not unblock"),
        Explorer(tmp_path, 0, "http://a.foo") as _,
    ):
        pass


@mark.parametrize("mode", (ExplorerMode.ALL, ExplorerMode.LOAD))
def test_explorer_modes(mocker, tmp_path, mode):
    """test Explorer()"""
    page_explorer = mocker.patch(
        "site_scout.explorer.PageExplorer", autospec=True
    ).return_value.__enter__.return_value
    page_explorer.current_url = "http://foo.foo"
    page_explorer.explore.return_value = State.CLOSED
    page_explorer.get.return_value = PageLoad.SUCCESS
    page_explorer.title = "title"
    with Explorer(tmp_path, 0, "http://foo.foo", mode=mode, pause=0) as explorer:
        assert not explorer._can_skip.is_set()
        # allow explore thread to complete
        explorer._thread.join(timeout=10)
        explorer.close()
        if mode == ExplorerMode.ALL:
            assert page_explorer.explore.call_count == 1
        elif mode == ExplorerMode.LOAD:
            assert page_explorer.explore.call_count == 0



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from logging import getLogger
from pathlib import Path
from platform import machine, system
from tempfile import TemporaryDirectory
from zipfile import ZIP_DEFLATED, ZipFile

from Collector.Collector import Collector
from FTB.ProgramConfiguration import ProgramConfiguration
from FTB.Signatures.CrashInfo import CrashInfo

LOG = getLogger(__name__)


class FuzzManagerReporter:
    """Report results to a FuzzManager server."""

    FM_CONFIG = Path.home() / ".fuzzmanagerconf"

    __slots__ = ("_conf", "_working_path")

    def __init__(
        self,
        binary: Path,
        fm_config: Path = FM_CONFIG,
        working_path: Path | None = None,
    ) -> None:
        self._working_path = working_path

        if fm_config and not fm_config.is_file():
            raise FileNotFoundError(f"Missing: {fm_config}")

        # create ProgramConfiguration that can be reported to a FM server
        if Path(f"{binary}.fuzzmanagerconf").is_file():
            # attempt to use "<binary>.fuzzmanagerconf"
            self._conf = ProgramConfiguration.fromBinary(binary)
        else:
            LOG.debug(
                "missing '%s.fuzzmanagerconf', creating ProgramConfiguration", binary
            )
            cpu = machine().lower()
            self._conf = ProgramConfiguration(
                binary.name,
                "x86_64" if cpu == "amd64" else cpu,
                system(),
            )

    @staticmethod
    def _read_ffpuppet_log(log_path: Path, log_id: str) -> list[str] | None:
        """Read logs created by FFPuppet.

        Args:
            log_path: Directory containing logs output by FFPuppet.
            log_id: Log to find. Must be: 'aux', 'stderr', 'stdout'.

        Returns:
            Lines from the requested log file.
        """
        assert log_id in {"aux", "stderr", "stdout"}, f"Invalid log id {log_id}"

        log_file: Path | None = None
        if log_id == "aux":
            # look for sanitizer logs
            for entry in log_path.glob("log_*.txt"):
                if "_asan_" in entry.name:
                    log_file = entry
                    break
            # look for minidump logs
            if not log_file:
                md_logs = sorted(log_path.glob("log_minidump_*.txt"))
                if md_logs:
                    log_file = md_logs[0]
            # TODO: valgrind logs
        else:
            log_file = log_path / f"log_{log_id}.txt"

        if log_file and log_file.is_file():
            # read text, sanitize and splitlines
            return (
                log_file.read_text("utf-8", errors="replace")
                .replace("\0", "?")
                .splitlines()
            )
        return None

    def submit(self, result: Path, metadata: dict[str, str]) -> tuple[int, str]:
        """Submit results to a FuzzManager server.

        Args:
            results: Directory containing logs to submit.
            metadata: Extra data to include in the report.

        Returns:
            Crash ID and the short signature.
        """
        collector = Collector(tool="site-scout")

        # TODO: add cache checks and report limits?

        # read the log files and create a CrashInfo object
        crash_info = CrashInfo.fromRawCrashData(
            self._read_ffpuppet_log(result, "stdout"),
            self._read_ffpuppet_log(result, "stderr"),
            self._conf,
            auxCrashData=self._read_ffpuppet_log(result, "aux"),
        )

        with TemporaryDirectory(prefix="fm-report", dir=self._working_path) as tmp_dir:
            # add result to a zip file
            zip_name = Path(tmp_dir) / f"{result.name}.zip"
            with ZipFile(zip_name, mode="w", compression=ZIP_DEFLATED) as zip_fp:
                # add test files
                for entry in result.rglob("*"):
                    if entry.is_file():
                        zip_fp.write(entry, arcname=entry.relative_to(result))

            # submit results to the FuzzManager server
            new_entry = collector.submit(
                crash_info, testCase=zip_name, testCaseQuality=5, metaData=metadata
            )

        return (int(new_entry["id"]), str(crash_info.createShortSignature()))
